#include <stdio.h>
#include <string.h>
#pragma hdrstop

/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.13.4, 2005-09-08, <e2500@narod.ru>

   Supported camera (tested)
      E2500
      E4500
      E5000
      E5900
      E995
      E885
      E775
      E990
      E950
      E900 (Thanks Sakura Shan)
      E2100
      E3100
      E3700
      E4300
      C4040Z
      C4000Z,C4100Z
      C2020Z
      C40Z,D40Zoom (Thanks Gustavo Jeronimo)
      Minolta Z2
   Main function

*/

#include <stdlib.h>
#include "tiff.h"
#include "file.h"
#include "raw2nef.h"

#define program "raw2nef v0.13.4, by paul69 <e2500@narod.ru> [compiled: " __DATE__ " " __TIME__ "]\n"

bool raw_to_nef(int n, char* a[]);

int main(int argc, char* argv[] )
{
   return raw_to_nef( argc, argv );
}

static void usage();
static char ofn_buf[_MAX_PATH];
static char efn_buf[_MAX_PATH];

static bool confirm_E2500(TCCDParam* ccd_pars, FILE* in)
{
   fseek( in, ccd_pars->file_size/2, 0 );
   unsigned size = 10*ccd_pars->row_length;
   unsigned test = 0;
   for(unsigned i=0; i<size; i += 3)
   {
      byte bitbuf[3];
      if( !fread( bitbuf, 3, 1, in) )
         return false;

      test |= (bitbuf[1]>>4)|(bitbuf[2]);
   }
   if( (test & 3) == 0 )
      return true;
   return false;
}

unsigned pack( const char* s )
{
   unsigned n = 0;
   for(int i=0; i<6; ++i)
   {
      unsigned char c = (unsigned char)*s++;
      unsigned d;

      if( c >= '0' && c <= '9' ) {
         d = c - '0';
      }
      else
      if( c >= 'A' && c <= 'Z' ) {
         d = 10 + c - 'A';
      }
      else
      if( c >= 'a' && c <= 'z' ) {
         d = 10 + c - 'a';
      }
      else
         break;

      n = n*36+d;
   }
   return n;
}

struct unpack_buff { char buf[12]; };
const char* unpack( unpack_buff& buf, unsigned x )
{
   char* s = &buf.buf[11];
   *(s) = 0;
   do
   {
      unsigned d = x % 36;
      *(--s) = (char)(d > 9 ? 'A'+d-10 : '0'+d);
      x = x / 36;
   } while( x );
   
   return s;
}

#define CAMERA_ID(id) pack(#id)

static
TCCDParam default_pars[] = {
   {1600+16, 1200+4, 0x2CE000, "YCGM", 2424, 12, CAMERA_ID(E2500), TCCDParam::fInterlaced, confirm_E2500 }, // E2500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, CAMERA_ID(E4500), TCCDParam::fInterlaced }, // E4500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, CAMERA_ID(C4040Z), TCCDParam::fInterlaced }, // Camedia C4040Z
   {   2304,   1716, 0x5A9000, "GMYC", 3456, 12, CAMERA_ID(C40Z),   TCCDParam::fInterlaced }, // Camedia C40Z,D40Zoom (Thanks Gustavo Jeronimo)
   {2560+16, 1920+4, 0x718000, "GMYC", 3864, 12, CAMERA_ID(E5000), TCCDParam::fInterlaced }, // E5000
   {2560+16, 1920+4, 0x718000, "GMYC", 3864, 12, CAMERA_ID(E5700), TCCDParam::fInterlaced }, // E5700
   {2048+16, 1536+4, 0x48D000, "MGCY", 3096, 12, CAMERA_ID(E995),  TCCDParam::fInterlaced }, // E995
   {2048+16, 1536+4, 0x48D000, "GMYC", 3096, 12, CAMERA_ID(E885),  TCCDParam::fInterlaced }, // E885
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10, CAMERA_ID(E775),  TCCDParam::fInterlaced }, // E775
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10, CAMERA_ID(C2020Z), TCCDParam::fInterlaced }, // Camedia C2020Z
   {2048+16, 1536+4, 0x48D000, "GMYC", 3096, 12, CAMERA_ID(E990),  TCCDParam::fInterlaced }, // E990
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10, CAMERA_ID(E950),  TCCDParam::fInterlaced }, // E950
   {1280+16,  960+4, 0x182004, "CYMG", 1632, 10, CAMERA_ID(E900),  TCCDParam::fInterlaced }, // E900

   {2288, 1710, 0x599000, "BGGR", 3432, 12, CAMERA_ID(E4300), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // E4300 (BGGR)
   {2288, 1710, 0x599000, "BGGR", 3432, 12, CAMERA_ID(Z2), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors|TCCDParam::fARM }, // Z2 (BGGR)
   {2064, 1542, 0x48E000, "GBRG", 3096, 12, CAMERA_ID(E3100), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors|TCCDParam::fARM }, // E3100 (GBRG)
   {1616, 1206, 0x2CE000, "RGGB", 2424, 12, CAMERA_ID(E2100), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors|TCCDParam::fARM }, // E2100 (RGGB) ++
   {2304, 1718, 0x5AA000, "BGGR", 3456, 12, CAMERA_ID(C4000Z), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // Olympus C4000Z (BGGR) ++
   {2304, 1718, 0x5AA000, "BGGR", 3456, 12, CAMERA_ID(C4100Z), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // Olympus C4100Z (BGGR) ++
//   {2260, 1700, 0x753FA0, "RGGB", 4520, 16, CAMERA_ID(QV4000), TCCDParam::fPrimaryColors }, // Casio QV-4000 (BGGR) KX821003.RAW
   {2048+16, 1538+4, 0x48E000, "RGGB", 3096, 12, CAMERA_ID(E3700), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors|TCCDParam::fARM }, // E3700 (RGGB)++
//   {2858, 0x6087C0/4288, 0x6087C0, "RGGB", 4288, 12, CAMERA_ID(EXP600), TCCDParam::fPrimaryColors }, // Casio EX-P600 (RGGB) 0103cimg.raw
   {2592+16, 1944+4, 0x747000, "GRBG", 3912, 12, CAMERA_ID(E5900), TCCDParam::fInterlaced|TCCDParam::fPrimaryColors|TCCDParam::fARM }, // E5900 (GRBG)++
};        

const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

bool Match( const char* FileName, const char* pattern )
{
   for(int i=0;; ++i)
   {
      if( !pattern[i] ) return true;
      
      if( pattern[i] == '#' )
      {
         if( FileName[i] >= '0' && FileName[i] <= '9' )
            continue;
         break;
      }
      if( pattern[i] >= 'A' && pattern[i] <= 'Z' )
      {
         if( (FileName[i] & ~0x20) == pattern[i] )
            continue;
         break;
      }
      
      if( FileName[i] == pattern[i] )
         continue;
      break;
   }
   return false;
}

char* CreateExifFileName(const char* FilePath, int ExtOffset)
{
   char* efn = efn_buf;
   memcpy(efn, FilePath, ExtOffset);
   memcpy(efn+ExtOffset, ".JPG", 5);

   char* num = efn+ExtOffset-1;
   for(int i=0; i<8; ++i)
   {
      int d = num[-i];
      if( d < '0' || d > '9' ) break;
      if( d != '9' )
      {
         num[-i] = (char)(d+1);
         break;
      }
      num[-i] = '0';
   }
   return efn_buf;
}

unsigned ParseExifCameraModel( const char* model )
{
   if( model[0] == 'E' || model[0] == 'C' )
      return pack( model );

   if( pack(model) == pack("DiMAGE") )
      return pack( model + 7 );

   return 0;
}

bool ProcessFile(const char* ifn, char* ofn, char* efn, unsigned camera,
   bool verbose, bool copy_exif, bool endian, bool compatible )
{
   int i;

   if( verbose )
      printf("input: %s\n", ifn);

   if( !ofn || !efn )
   {
      const char* inpFileName = ifn;

      const char* ifn_slash = strrchr(inpFileName, '/');
      if( ifn_slash )
         inpFileName = ifn_slash+1;

      const char* ifn_bslash = strrchr(inpFileName, '\\');
      if( ifn_bslash )
         inpFileName = ifn_bslash+1;

      const char* inpFileExt = strrchr(inpFileName, '.');

      if( !inpFileExt )
         inpFileExt = ifn + strlen(ifn);

      if( !ofn )
      {
         ofn = ofn_buf;
         int len = inpFileExt - ifn;
         memcpy(ofn, ifn, len);
         strcpy(ofn+len, ".NEF");
      }

      if( !efn && copy_exif )
      {
         efn = CreateExifFileName(ifn, inpFileExt-ifn);
         if( !efn && verbose )
            printf("warning: cannot get exif filename\n");
      }
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - cannot open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   NEF_IFD Nef_IFD;

   if( efn )
   {
      if( verbose )
         printf("read exif info from \"%s\"\n", efn);

      TFile exif = fopen( efn, "rb");

      if( !exif && verbose )
         { printf("warning: \"%s\" - cannot open exif file\n", efn); }
      else
      if( !Nef_IFD.read_exif( exif ) && verbose )
         { printf("warning: \"%s\" - cannot read exif file\n", efn); }
   }

   IFDir::Tag* Model = Nef_IFD.IFD0.get_tag( TIFF::Model );
   if( Model )
   {
      unsigned exif_camera = ParseExifCameraModel( (const char*)Model->value );

      if( exif_camera != 0 )
      {
         if( camera == 0 )
            camera = exif_camera;
         else
         if( camera != exif_camera )
         {  unpack_buff buf1, buf2;
            printf("warning: exif model %s and option %s are different\n", unpack(buf1, exif_camera), unpack(buf2, camera) ); }
      }
   }

   for(i=0; i<n_default_pars; ++i)
   {
      if( camera == 0 )
      {
         if( file_size == default_pars[i].file_size )
         {
            if( !default_pars[i].confirm )
               break;

            if( default_pars[i].confirm(&default_pars[i], in) )
               break;
         }
      }
      else
      {
         if( camera == default_pars[i].camera )
            break;
      }
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      unpack_buff buf;
      if( verbose )
         printf("camera model is %s\n", unpack( buf, default_pars[i].camera) );

      ccd_pars = default_pars[i];
   }
   else
   {
      printf("error: unsupported format\n");
      return false;
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read raw data from \"%s\"\n", ifn);

   if( !read_raw( in, ccd, ccd_pars) )
   {
      printf("error: can not read raw data\n");
      return false;
   }

   in.close();

   if( verbose )
      printf("write picture to: \"%s\"\n", ofn);

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( !Nef_IFD.write_nef(out, endian, ccd, ccd_pars, compatible) )
      { printf("error: can not write NEF\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}


bool raw_to_nef(int n, char* a[])
{
   // endian: true for Intel, false for Motorolla
   bool endian = true;
   bool compatible = false;
   bool copy_exif = true;
   char* ifn = NULL;
   int i_count = 0;
   char* ofn = NULL;
   char* efn = NULL;
   bool verbose = false;
   bool batch_mode = true;
   unsigned camera = 0;

   if( n <= 1 )
   {
      usage();
      return false;
   }

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'v': verbose = (a[i][2] != '-'); break;
         case 'c': compatible = (a[i][2] != '-'); break;

         case 'e':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               efn = a[++i];
            else
               goto missing;
            break;

         case 'o':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               ofn = a[++i];
            else
               goto missing;
            break;

         case '?': usage(); return false;

         case 'a': camera = 0; break;

         default:
            if( a[i][1] >= '1' && a[i][1] <= '9' )
            {
               a[i][0] = 'E';
               camera = pack( a[i] );
            }
            else
            if( (a[i][1] == 'E' || a[i][1] == 'E' || a[i][1] == 'Z') && a[i][2] >= '1' && a[i][2] <= '9' )
               camera = pack( a[i]+1 );
            else
               printf("warning: unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("warning: invalid switch: %s\n", a[i]);
            break;

         missing:
            printf("warning: missing argument for %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
            
         ++i_count;
      }
   }

   if( verbose )
      printf( program );

   if( !ifn )
   {
      printf("error: input file is missing\n\n");
      return false;
   }

   if( batch_mode && i_count > 1 )
   {
      for(i=1; i<n; ++i)
      {
         if( a[i][0] == '-' )
         {
            switch( a[i][1] )
            {
            case 'o':
            case 'e':
               if( i+1 < n && a[i+1][0] != '-' )
                  ++i;
               break;
            }
         }
         else
         {
            if( a[i][0] )
            {
               if( !ProcessFile(a[i], NULL, NULL, camera, verbose, copy_exif, endian, compatible) )
                  printf("cannot process \"%s\"\n\n", a[i]);
            }
         }
      }
   }
   else
   {
      if( !ProcessFile(ifn, ofn, efn, camera, verbose, copy_exif, endian, compatible) )
      {
         printf("cannot process \"%s\"\n", ifn);
         return false;
      }
   }
   return true;
}

void usage()
{
   printf("Coolpix RAW data to E5700/E8700 NEF file converter\n");
   printf( program );

   printf("\nargs: [options] input_file1 [input_file2...]\n");

   printf("\noptions:\n");
   printf("   -auto - autodetect (default)\n");
   printf("   -E2100 - Coolpix 2100\t");  printf("   -C4100Z - Camedia C4000Z/C4100Z\n");
   printf("   -E2500 - Coolpix 2500\t");  printf("   -C4040Z - Camedia C4040Z\n");
   printf("   -E3100 - Coolpix 3100\t");  printf("   -C2020Z - Camedia C2020Z\n");
   printf("   -E3700 - Coolpix 3700*\t"); printf("   -C40Z - Camedia C40Z/D40Z\n");
   printf("   -E4300 - Coolpix 4300\t");  printf("   -E900 - Coolpix 900\n");
   printf("   -E4500 - Coolpix 4500\t");  printf("   -E950 - Coolpix 950\n");
   printf("   -E5000 - Coolpix 5000\t");  printf("   -E990 - Coolpix 990*\n");
   printf("   -E5700 - Coolpix 5700\t");  printf("   -E775 - Coolpix 775\n");
   printf("   -E5900 - Coolpix 5900\n");
   printf("   -E995 - Coolpix 995\t");    printf("   -E885 - Coolpix 885*\n");
   printf("   -Z2 - Minolta Z2*\n");    
   printf("   (* auto is wrong for this camera!)\n");

   printf("\n");
   printf("   -v - verbose\n");
   printf("   -c - compatible with E5700/E8700\n");
   printf("   -e <filename> - exif jpeg file name (disable batch mode)\n");
   printf("   -o <filename> - nef output file name (disable batch mode)\n");

   printf("\nSupported cameras: " );
   printf("Nikon Coolpix: E2100, E2500, E3100, E3700, E4300, E4500, E5000, E5900, E995, E885, E775, E990, E950, E900");
   printf(" / Olympus Camedia: C4000Z, C4100Z, C4040Z, C40Z/D40Z, C2020Z / Minolta Z2");
}

